/* TIGRE Graph reducer:  REDUCE.C */
/* Driver code */

/* (C) Copyright 1989   Philip Koopman Jr.  */
/*  Last update:  4/17/89   */
/*  Test code version -- not for distribution  */

#include "reduce.h"
#include "heap.h"
#include "lexer.h"
#include "sim.h"

void dumpgraph(Celltype *root, int indent);

   Celltype *temp_pointer ;

   Valuetype ip ;

   Celltype underflow_node[10] ;   /* used for underflow trapping */

   long count;   /* counts number of combinators used */
   int error_flag ;  /* flag if graph has an error */

   token nodetok  ;  /* used as token input buffer */
   Celltype do_lit_value,do_i_value ; /* save DO_LIT and DO_I values */

   long i_counts[MAX_TOKEN+10]; /* combinator freq. counts */
   long i_heap   ;  /* number of heap nodes allocated */

  int       ready_for_longjmp ;
  jmp_buf   env;

#define INDENT_SIZE 1

#if UNIX
/* variables for time measurement: */
struct tms      time_info;
extern long     times (struct tms *time_info);
long            Begin_Time, End_Time;
#else
clock_t         Begin_Time, End_Time;
#endif
float           Seconds, Combinators_Per_Second;

/* ______________________________________________________________________ */
void spine_under(char * xx)
  { printf("\n\nspine stack underflow  %s\n",xx);
    /* xx is passed to this function in part to trick the compiler */
    exit(-1);
  }

void spine_overflow()
  { printf("\n\nspine stack overflow\n");
    exit(-1);
  }
/* ______________________________________________________________________ */

void dumphalf(Celltype *cell,int indent)
  {  int i;
     printf("\n");
     for ( i = 0 ; i <= indent ; i++ )
         printf(" ");
#if UNIX
     printf("%d) %d %X: ",indent/INDENT_SIZE,
               ((cell-(heap_start+HEAP_STRIDE+1))/HEAP_STRIDE),cell);
#else
     printf("%d) %ld %p: ",indent/INDENT_SIZE,
               ((cell-(heap_start+HEAP_STRIDE+1))/HEAP_STRIDE),cell);
#endif
     if ( IS_PTR(cell->value.child) )
#if UNIX
         { printf("Call: %X",UNFUDGE(cell->value.child)); }
#else
         { printf("Call: %p",UNFUDGE(cell->value.child)); }
#endif
     else
         {  if (cell->value.comb == mapval(DO_I))  printf("I");
            else if (cell->value.comb == mapval(DO_S))  printf("S");
            else if (cell->value.comb == mapval(DO_SPRIME))  printf("S'");
            else if (cell->value.comb == mapval(DO_K))  printf("K");
            else if (cell->value.comb == mapval(DO_B))  printf("B");
            else if (cell->value.comb == mapval(DO_BSTAR))  printf("B*");
            else if (cell->value.comb == mapval(DO_C))  printf("C");
            else if (cell->value.comb == mapval(DO_CPRIME)) printf("C'");
            else if (cell->value.comb == mapval(DO_LIT))  printf("LIT");
            else if (cell->value.comb == mapval(DO_ZERO))  printf("0");
            else if (cell->value.comb == mapval(DO_ONE))  printf("1");
            else if (cell->value.comb == mapval(DO_TWO))  printf("2");
            else if (cell->value.comb == mapval(DO_THREE))  printf("3");
            else if (cell->value.comb == mapval(DO_PLUS))  printf("+");
            else if (cell->value.comb == mapval(DO_MINUS))  printf("-");
            else if (cell->value.comb == mapval(DO_TIMES))  printf("*");
            else if (cell->value.comb == mapval(DO_DIV))  printf("/");
            else if (cell->value.comb == mapval(DO_MOD))  printf("MOD");
            else if (cell->value.comb == mapval(DO_LESS))  printf("<");
            else if (cell->value.comb == mapval(DO_GREAT))  printf(">");
            else if (cell->value.comb == mapval(DO_IF))  printf("IF");
            else if (cell->value.comb == mapval(DO_EQ))  printf("=");
            else if (cell->value.comb == mapval(DO_NOT))  printf("NOT");
            else if (cell->value.comb == mapval(DO_ABS))  printf("ABS");
            else if (cell->value.comb == mapval(DO_TRUE))  printf("TRUE");
            else if (cell->value.comb == mapval(DO_FALSE))  printf("FALSE");
            else if (cell->value.comb == mapval(DO_NIL))  printf("NIL");
            else if (cell->value.comb == mapval(DO_PRINT))  printf("PRINT");
            else if (cell->value.comb == mapval(DO_NL))  printf("NL");
            else if (cell->value.comb == mapval(DO_P))  printf("P");
            else if (cell->value.comb == mapval(DO_U))  printf("U");
            else if (cell->value.comb == mapval(DO_AND))  printf("AND");
            else if (cell->value.comb == mapval(DO_OR))  printf("OR");
            else if (cell->value.comb == mapval(DO_S1))  printf("$1");
            else if (cell->value.comb == mapval(DO_S2))  printf("$2");
            else if (cell->value.comb == mapval(DO_S3))  printf("$3");
            else if (cell->value.comb == mapval(DO_S4))  printf("$4");
            else if (cell->value.comb == mapval(DO_S5))  printf("$5");
            else if (cell->value.comb == mapval(DO_S6))  printf("$6");
            else if (cell->value.comb == mapval(DO_S7))  printf("$7");
            else if (cell->value.comb == mapval(DO_S8))  printf("$8");
            else if (cell->value.comb == mapval(DO_FLOOR))  printf("/_");
            else if (cell->value.comb == mapval(DO_S10))  printf("$10");
            else if (cell->value.comb == mapval(DO_S11))  printf("$11");
            else if (cell->value.comb == mapval(DO_S12))  printf("$12");
            else if (cell->value.comb == mapval(DO_S13))  printf("$13");
            else if (cell->value.comb == mapval(DO_S14))  printf("$14");
            else if (cell->value.comb == mapval(DO_S15))  printf("$15");
            else if (cell->value.comb == mapval(DO_S16))  printf("$16");
            else if (cell->value.comb == mapval(DO_S17))  printf("$17");
            else if (cell->value.comb == mapval(DO_S18))  printf("$18");
            else if (cell->value.comb == mapval(DO_S19))  printf("$19");
            else if (cell->value.comb == mapval(NOT_A_TOKEN))  printf("????");
            else  printf("\n\n**Bad Node!! %X %X\n",cell->value.comb,mapval(DO_S));
         } ;
  }
/* ______________________________________________________________________ */

void dumpgraph(Celltype *root, int indent)
  {  if (root < heap_start || root > heap_end )
       { printf("\n*** node out of range at address %X\n",root);
         return;
       }
     if ( IS_FORWARDED(root->value.child) )
     { printf(" ==> sharing/recursion");
       return ;
     }
     dumphalf(root,indent);
     if ( (INT_32) (root->value.comb) == mapval(DO_LIT))
        {
#if UNIX
          printf("  value=%d",(root+1)->value.literal);
#else
          printf("  value=%ld",(root+1)->value.literal);
#endif
          root->value.child = MAKE_FORWARD(root->value.child) ;
          return ;
        }
     root->value.child = MAKE_FORWARD(root->value.child) ;
     if ( IS_PTR(root->value.child) )
         dumpgraph(UNFUDGE(UNFORWARD(root->value.child)),indent+INDENT_SIZE);
     dumphalf((root+1),indent);
     if ( IS_PTR( (root+1)->value.child) )
         dumpgraph(UNFUDGE((root+1)->value.child),indent+INDENT_SIZE);
  }
/* ______________________________________________________________________ */

#define INST_ARG(x)  i_counts[x],  \
      ( 100.0 * ( (float) i_counts[x] )) / ( (float) count )

#define COMB_ARG(x)  (x), ( ((float) (x) ) / ((float) count) )

void dump_instruments()
  {
     printf("\n");
     printf("\nHeap nodes allocated = %ld  (%5.3f/comb)\n",
          COMB_ARG(i_heap));
     printf("\nStack nodes unwound  = %ld  (%5.3f/comb)\n",
          COMB_ARG(i_counts[DO_UNWIND]));
     printf("\n");
     printf("\nData access statistics:");
     printf("\n # double word data accesses  = %7ld  (%5.3f/comb)\n",
          COMB_ARG(count_double)) ;
     printf("\n");
     printf("\n# heap reads         = %7ld  (%5.3f/comb)",
          COMB_ARG(count_data_read));
     printf("\n# heap writes        = %7ld  (%5.3f/comb)",
          COMB_ARG(count_data_write));
     printf("\n                 # total heap accesses  = %7ld  (%5.3f/comb)",
          COMB_ARG(count_data_read+count_data_write));
     printf("\n# top of stack read  = %7ld  (%5.3f/comb)",
          COMB_ARG(count_stackme_read));
     printf("\n# top of stack write = %7ld  (%5.3f/comb)",
          COMB_ARG(count_stackme_write));
     printf("\n# 2nd of stack read  = %7ld  (%5.3f/comb)",
          COMB_ARG(count_stackp_read));
     printf("\n# 2nd of stack write = %7ld  (%5.3f/comb)",
          COMB_ARG(count_stackp_write));
     printf("\n# 3rd of stack read  = %7ld", count_stackg_read);
     printf("\n# 3rd of stack write = %7ld", count_stackg_write);
     printf("\n                 # total stack reads    = %7ld  (%5.3f/comb)",
          COMB_ARG(count_stackme_read +
                   count_stackp_read + count_stackg_read));
     printf("\n                 # total stack writes   = %7ld  (%5.3f/comb)",
          COMB_ARG(count_stackme_write +
                   count_stackp_write + count_stackg_write));
     printf("\n                 # total stack accesses = %7ld  (%5.3f/comb)",
          COMB_ARG(count_stackme_read  + count_stackp_read  +
                   count_stackg_read   + count_stackme_write +
                   count_stackp_write + count_stackg_write));
     printf("\n");
     printf("\n                 # total data reads     = %7ld  (%5.3f/comb)",
          COMB_ARG( count_data_read     +
                    count_stackme_read  + count_stackp_read  +
                    count_stackg_read));
     printf("\n                 # total data writes    = %7ld  (%5.3f/comb)",
          COMB_ARG( count_data_write   +
                    count_stackme_write + count_stackp_write +
                    count_stackg_write));
     printf("\n                 # total data accesses  = %7ld  (%5.3f/comb)",
          COMB_ARG( count_data_read     + count_data_write   +
                    count_stackme_read  + count_stackp_read  +
                    count_stackg_read +
                    count_stackme_write + count_stackp_write +
                    count_stackg_write));
     printf("\n");
     printf("\nCombinator frequencies:");
     printf("\n       I %6ld    %6.3f%%",INST_ARG(DO_I  ));
     printf("\n       K %6ld    %6.3f%%",INST_ARG(DO_K  ));
     printf("\n       S %6ld    %6.3f%%",INST_ARG(DO_S  ));
     printf("\n      S' %6ld    %6.3f%%",INST_ARG(DO_SPRIME));
     printf("\n       B %6ld    %6.3f%%",INST_ARG(DO_B));
     printf("\n      B* %6ld    %6.3f%%",INST_ARG(DO_BSTAR));
     printf("\n       C %6ld    %6.3f%%",INST_ARG(DO_C));
     printf("\n      C' %6ld    %6.3f%%",INST_ARG(DO_CPRIME));
     printf("\n       + %6ld    %6.3f%%",INST_ARG(DO_PLUS));
     printf("\n       - %6ld    %6.3f%%",INST_ARG(DO_MINUS));
     printf("\n       * %6ld    %6.3f%%",INST_ARG(DO_TIMES));
     printf("\n       / %6ld    %6.3f%%",INST_ARG(DO_DIV));
     printf("\n      /_ %6ld    %6.3f%%",INST_ARG(DO_FLOOR));
     printf("\n     MOD %6ld    %6.3f%%",INST_ARG(DO_MOD));
     printf("\n     NOT %6ld    %6.3f%%",INST_ARG(DO_NOT));
     printf("\n     ABS %6ld    %6.3f%%",INST_ARG(DO_ABS));

     printf("\n      IF %6ld    %6.3f%%",INST_ARG(DO_IF));
     printf("\n       < %6ld    %6.3f%%",INST_ARG(DO_LESS));
     printf("\n       = %6ld    %6.3f%%",INST_ARG(DO_EQ));
     printf("\n       > %6ld    %6.3f%%",INST_ARG(DO_GREAT));
     printf("\n      OR %6ld    %6.3f%%",INST_ARG(DO_OR));
     printf("\n     AND %6ld    %6.3f%%",INST_ARG(DO_AND));

     printf("\n     LIT %6ld    %6.3f%%",INST_ARG(DO_LIT));
     printf("\n       0 %6ld    %6.3f%%",INST_ARG(DO_ZERO));
     printf("\n       1 %6ld    %6.3f%%",INST_ARG(DO_ONE));
     printf("\n       2 %6ld    %6.3f%%",INST_ARG(DO_TWO));
     printf("\n       3 %6ld    %6.3f%%",INST_ARG(DO_THREE));
     printf("\n    TRUE %6ld    %6.3f%%",INST_ARG(DO_TRUE));
     printf("\n   FALSE %6ld    %6.3f%%",INST_ARG(DO_FALSE));

     printf("\n       P %6ld    %6.3f%%",INST_ARG(DO_P));
     printf("\n       U %6ld    %6.3f%%",INST_ARG(DO_U));
     printf("\n     NIL %6ld    %6.3f%%",INST_ARG(DO_NIL));
     printf("\n   PRINT %6ld    %6.3f%%",INST_ARG(DO_PRINT));
     printf("\n      NL %6ld    %6.3f%%",INST_ARG(DO_NL));
     printf("\n\n");
     printf("\n      $1 %6ld    %6.3f%%",INST_ARG(DO_S1));
     printf("\n      $2 %6ld    %6.3f%%",INST_ARG(DO_S2));
     printf("\n      $3 %6ld    %6.3f%%",INST_ARG(DO_S3));
     printf("\n      $4 %6ld    %6.3f%%",INST_ARG(DO_S4));
     printf("\n      $5 %6ld    %6.3f%%",INST_ARG(DO_S5));
     printf("\n      $6 %6ld    %6.3f%%",INST_ARG(DO_S6));
     printf("\n      $7 %6ld    %6.3f%%",INST_ARG(DO_S7));
     printf("\n      $8 %6ld    %6.3f%%",INST_ARG(DO_S8));
     printf("\n     $10 %6ld    %6.3f%%",INST_ARG(DO_S10));
     printf("\n     $11 %6ld    %6.3f%%",INST_ARG(DO_S11));
     printf("\n     $12 %6ld    %6.3f%%",INST_ARG(DO_S12));
     printf("\n     $13 %6ld    %6.3f%%",INST_ARG(DO_S13));
     printf("\n     $14 %6ld    %6.3f%%",INST_ARG(DO_S14));
     printf("\n     $15 %6ld    %6.3f%%",INST_ARG(DO_S15));
     printf("\n     $16 %6ld    %6.3f%%",INST_ARG(DO_S16));
     printf("\n     $17 %6ld    %6.3f%%",INST_ARG(DO_S17));
     printf("\n     $18 %6ld    %6.3f%%",INST_ARG(DO_S18));
     printf("\n     $19 %6ld    %6.3f%%",INST_ARG(DO_S19));
     printf("\n\n");
  }

/* ______________________________________________________________________ */

/* WORDS TO HELP BUILD THE TREE */

Celltype *makelit(token *tok)
  { Celltype *temp1 ;
    temp1 = newnode(spine_stack);
    Ltemp1.comb = mapval(DO_LIT) ;
    Rtemp1.literal = tok->value ;
    return(temp1);
  }
/* ______________________________________________________________________ */

Celltype *node()
  {  /* returns pointer to a node */
     Celltype *temp1 ;
     token tok ;
     temp1 = newnode(spine_stack);  /* allocate the new node */
     /* get left side */
     GetNextToken(&tok);
     if(tok.toktype != LPAREN) Ltemp1.comb = mapval(tok.toktype);
     if(tok.toktype == LPAREN)
             { Ltemp1.child = FUDGE(node()); /* points to subtree */
             }
     if(tok.toktype == DO_LIT)
             { Ltemp1.child = FUDGE(makelit(&tok)); /* make a literal */
             }
     if(tok.toktype == CYCLE)
       { /* push patch address of node onto spine stack */
         *(--spine_stack) = temp1 ;
       }
     /* get right side */
     GetNextToken(&tok);
     if(tok.toktype != LPAREN) Rtemp1.comb = mapval(tok.toktype);
     if(tok.toktype == LPAREN)
             { Rtemp1.child = FUDGE(node()); /* points to subtree */
             }
     if(tok.toktype == DO_LIT)
             { Rtemp1.child = FUDGE(makelit(&tok)); /* make a literal */
             }
     if(tok.toktype == CYCLE)
       { /* push patch address of node onto spine stack */
         *(--spine_stack) = temp1+1 ;
       }
     /* get the right parenthesis */
     GetNextToken(&tok);
    if(tok.toktype != RPAREN)  printf(" ** missing ')' inserted** ");
     return(temp1);
   }
/* ______________________________________________________________________ */

Celltype *SetupFunction(INT_32 param)
  { Celltype *temp1,*temp2 ;
    temp1 = node();
    /* fix up cycle pointers */
    while (spine_stack < spine_end )
       { temp2 = *(spine_stack++) ;
         temp2->value.child = FUDGE(temp1) ;
       }
    /* glue on parameter */
    nodetok.value = param;
    temp2 = newnode(spine_stack);  /* allocate the root node */
    Ltemp2.child = FUDGE(temp1);
    Rtemp2.child = FUDGE(makelit(&nodetok));
    return(temp2) ;
  }
/* ______________________________________________________________________ */

void SetTripleCell(Celltype *temp1, token tok)
  {
    switch(tok.toktype)
    { case DO_LIT:
        temp1->value.child = FUDGE(makelit(&tok));
        break ;
      case DO_NODE:
        temp1->value.child =
                   FUDGE(heap_start+(tok.value*HEAP_STRIDE)+HEAP_STRIDE+1) ;
        break ;
      case NOT_A_TOKEN:  case LPAREN:  case RPAREN:  case EOF_TOKEN:
        printf("\nBad input file\n");
        exit(-1);
        break ;
      default:  /* combinators */
        temp1->value.comb = mapval(tok.toktype) ;
    }
    return;
  }
/* ______________________________________________________________________ */

Celltype *TripleFunction(INT_32 param)
  { token lefttok, righttok ;
    Celltype *temp1,*temp2 ;
    while ((nodetok.toktype != EOF_TOKEN) && (error_flag == FALSE) )
     {
      if(nodetok.toktype == DO_ZERO)   nodetok.toktype = DO_LIT ;
      if(nodetok.toktype == DO_ONE)    nodetok.toktype = DO_LIT ;
      if(nodetok.toktype == DO_TWO)    nodetok.toktype = DO_LIT ;
      if(nodetok.toktype == DO_THREE)    nodetok.toktype = DO_LIT ;

      if(nodetok.toktype != DO_LIT) { printf("\nBad input file\n"); exit(-1); }
      GetNextToken(&lefttok);
      GetNextToken(&righttok);
#if DEBUG
      printf(" (node %X) ",nodetok.value);
#endif
      temp1 = heap_start+(nodetok.value*HEAP_STRIDE)+HEAP_STRIDE+1;
                    /* allocate the new node */
      SetTripleCell(temp1,lefttok);
      SetTripleCell((temp1+1),righttok);
#if DEBUG
      dumphalf(temp1,13);
      dumphalf(temp1+1,13);
#endif
     GetNextToken(&nodetok);  /* get first element of next triple */
     }

    /* glue on parameter */
    nodetok.value = param;
    temp2 = newnode(spine_stack);  /* allocate the root node */
    Ltemp2.child = FUDGE(heap_start+HEAP_STRIDE+1) ;
                          /* root of input tree always node 0 */
    Rtemp2.child = FUDGE(makelit(&nodetok));
    return(temp2) ;
  }
/* ______________________________________________________________________ */

Celltype *ReadFunction(INT_32 param)
  {  Celltype *temp1 ;
    spine_stack = spine_end ;  /* holds patch points for cycles */
    error_flag = FALSE ;
    init_heap();
    GetNextToken(&nodetok);
    if(nodetok.toktype == LPAREN)
           { root_save = SetupFunction(param);  }
    else   { root_save = TripleFunction(param); }
     /* set up special node for trapping underflows mapval( */
    temp1 = to_heap + 1;
    Ltemp1.comb = mapval(SPINE_UNDERFLOW) ;
    Rtemp1.comb = mapval(SPINE_UNDERFLOW) ;
    temp1 = from_heap + 1 ;
    Ltemp1.comb = mapval(SPINE_UNDERFLOW) ;
    Rtemp1.comb = mapval(SPINE_UNDERFLOW) ;
    spine_stack = &spine_mem[SPINESIZE] ;
    *(--spine_stack) = temp1+1 ;
    *(--spine_stack) = temp1+1 ;
    *(--spine_stack) = temp1+1 ;
    *(--spine_stack) = temp1+1 ;
    spine_stack = spine_end ;
    collect_garbage(spine_stack);
    return(root_save);
 }
/* ______________________________________________________________________ */

void main(int argc,char *argv[])
  { Valuetype result ;
   int inparam;
   Celltype **spine_save;

   if(argc < 2)
     { printf("\nno file name given\n");
       return;
     }

#if !DUMP_TRACE
     printf("\nTIGRE graph reducer, ");
#if (! UNIX) && VAX_ASM
     printf(" MACRO ERROR #1 );
#endif
#if (! UNIX) && MIPS_ASM
     printf(" MACRO ERROR #1 );
#endif
#if VAX_ASM
     printf("VAX assembler version.");
#endif
#if MIPS_ASM
     printf("MIPS R2000 assembler version.");
#endif
#if ( ! MIPS_ASM ) && ( ! VAX_ASM )
     printf("C version.");
#endif
   printf("\nHeap size = %d nodes, ",HEAPSIZE);
#if ! VAX_ASM
   printf("Stack size = %d nodes, ",SPINESIZE);
#endif
   printf("input file: %s\n",argv[1]);
#endif

   from_heap =
          (Celltype *) malloc(sizeof(Celltype)*(HEAPSIZE+64)*HEAP_STRIDE) ;
   to_heap =
          (Celltype *) malloc(sizeof(Celltype)*(HEAPSIZE+64)*HEAP_STRIDE) ;
/*   spine_mem =
          (Celltype *) malloc(4 * SPINE_SIZE) ; */
   /* force heaps to align on octaword-byte boundary */
   from_heap = (Celltype *) ( ((((INT_32) from_heap)+31) & -32)  ) ;
   to_heap = (Celltype *) ( ((((INT_32) to_heap)+31) & -32)  ) ;
   do_lit_value.value.comb = mapval(DO_LIT) ;
   do_i_value.value.comb = mapval(DO_I) ;

   inparam = 12345 ;
   while ( inparam > 0 )
   {
#if ! DUMP_TRACE
     printf("\nInput value (0 to stop, -1 to dump graph)?");
#endif
     scanf("%d",&inparam);
#if ! DUMP_TRACE
     printf("input = %d\n",inparam);
#endif
     if(feof(stdin))  return ;
     if(inparam == 0) return ;

     InitLexer(argv[1]);
     root_save = ReadFunction(inparam);
     EndLexer();
     if(inparam == -1 )
     { printf("\n\n ***** graph dump *****");
       zapflags();
       dumpgraph(root_save,0);
       zapflags();
       printf("\n ***** end of graph dump *****\n");
       inparam = 1 ;
       continue ;
     }
     if(error_flag)
       { printf("\nill-formed graph\n");
         return;
       }

     count = 0 ;
     gc = 0 ;
     Initialize_Instruments();

#if UNIX
       /* Start timer */
       times (&time_info);
       Begin_Time = (long) time_info.tms_utime;
#else
       Begin_Time = clock();
#endif
       /* Evaluate graph */

       spine_save = spine_stack ;
#if UNIX
       _setjmp(env);
#else
       setjmp(env);
#endif
       ready_for_longjmp = TRUE;
#if DEBUG_TRACE
    printf("calling root node\n");
#endif
       ip.ptr = FUDGE(root_save) ;
       result = Evaluate(ip,spine_save);
       ready_for_longjmp = FALSE;

#if UNIX
       /* Stop timer */
       times (&time_info);
       End_Time = (long) time_info.tms_utime;
#else
       End_Time = clock();
#endif
#if ! DUMP_TRACE
       printf("And the answer is.....%d   ",result.literal);
       printf("(using %ld combinators and %d GCs)",count,gc);
#endif

#if (! INSTRUMENTS) && (!DUMP_TRACE)
       /* compute elapsed time */
       Seconds = ((float) (End_Time - Begin_Time) ) / HZ ;
       if (Seconds < 0.5)
       { printf("\nElapsed time < 0.5 seconds");
       } else {
         Combinators_Per_Second = count / Seconds ;
         printf ("\nElapsed time in seconds: %8.3f  ", Seconds);
         printf ("Combinators per Second: %8.0f",Combinators_Per_Second);
       }
       printf ("\n");
#endif

#if DUMP_TRACE
       printf("4 0   flush the cache\n");
#endif

#if INSTRUMENTS
       dump_instruments();
#endif
   }
   return;
  }
