/* TIGRE implementation:  SIM.C */
/* simulation hooks for cache, etc. simulations */
/* tailored to MIPS R2000 assembler language implementation */
 
/* (C) Copyright 1989   Philip Koopman Jr.  */
/*  Last update:  2/28/89   */
/*  Test code version -- not for distribution  */
 
#include "reduce.h"
#include "heap.h"
#include "sim.h"
 
long count_program ; /* i-fetches = instructions executed */
long count_data_read    ; /* data fetches */
long count_data_write   ; /* data stores */
long count_stackme_read ; /* top of stack reads */
long count_stackme_write ; /* top of stack writes */
long count_stackp_read  ; /* second on stack reads */
long count_stackp_write  ; /* second on stack writes */
long count_stackg_read  ; /* third on stack reads */
long count_stackg_write  ; /* third on stack writes */
long count_double        ; /* double word accesses */
 
void Initialize_Instruments()
{
  int i ;  /* loop counter */
  /* initialize combinator frequency counts */
  for ( i = 0 ; i <= MAX_TOKEN ; i++ )
    { i_counts[i] = 0 ; }
  i_heap = 0 ;
 
  /* initialize memory access counts */
  count_program       = 0 ; /* i-fetches = instructions executed */
  count_data_read     = 0 ; /* data fetches */
  count_data_write    = 0 ; /* data stores */
  count_stackme_read  = 0 ; /* top of stack read */
  count_stackme_write = 0 ; /* top of stack writes */
  count_stackp_read   = 0 ; /* second on stack reads */
  count_stackp_write  = 0 ; /* second on stack writes */
  count_stackg_read   = 0 ; /* third on stack reads */
  count_stackg_write  = 0 ; /* third on stack writes */
  count_double        = 0 ;
}
 
void simulate(space,  daddr )
Memory_Space space; INT_32 daddr;
{ /* space is memory space enum for the data access
   * daddr is the data address
   */
  int read_flag ;
 
   if( daddr == 0 )
      printf("\nWarning: access to address 0!");
 
#if DUMP_TRACE
   if (count_data_read > 1000000)  exit(-1);
#endif
 
   /* switch depending on memory space */
   switch( space )
   {
     case rd:
       count_data_read ++ ;
       read_flag = TRUE;
       break ;
 
     case wr:
       count_data_write ++ ;
       read_flag = FALSE;
       break ;
 
     case rdme:
       count_stackme_read ++ ;
       read_flag = TRUE;
       break ;
 
     case wrme:
       count_stackme_write ++ ;
       read_flag = FALSE;
       break ;
 
     case rdp:
       count_stackp_read ++ ;
       read_flag = TRUE;
       break ;
 
     case wrp:
       count_stackp_write ++ ;
       read_flag = FALSE;
       break ;
 
     case rdg:
       count_stackg_read ++ ;
       read_flag = TRUE;
       break ;
 
     case wrg:
       count_stackg_write ++ ;
       read_flag = FALSE;
       break ;
 
     default:
       printf("\nWarning: memory space=none");
       break ;
   }
#if ! DUMP_TRACE
  read_flag = read_flag ;
#endif
#if UNIX && DUMP_TRACE
   if ( read_flag )
     { printf("0 %x\n", daddr); }
   else
     { printf("1 %x\n", daddr); }
#endif
#if (! UNIX ) && DUMP_TRACE
   if ( read_flag )
     { printf("0 %lx\n", daddr); }
   else
     { printf("1 %lx\n", daddr); }
#endif
}
 
